/**
 * Copyright (c) 2017-present, Facebook, Inc.
 * All rights reserved.
 *
 * This source code is licensed under the BSD-style license found in the
 * LICENSE file in the root directory of this source tree.
 */

#pragma once

#include <string>

namespace gloo {

std::string getHostname();

bool useRankAsSeqNumber();

bool isStoreExtendedApiEnabled();

bool disableConnectionRetries();

} // namespace gloo
